/*
 *  linux/arch/ppc/kernel/setup.c
 *
 *  TCD version
 *    Copyright (C) 1998 Teleworld, Inc. All Rights Reserved.
 *
 *  PowerPC version
 *    Copyright (C) 1995-1996 Gary Thomas (gdt@linuxppc.org)
 *
 *  Adapted for Power Macintosh by Paul Mackerras
 *    Copyright (C) 1996 Paul Mackerras (paulus@cs.anu.edu.au)
 *
 *  Derived from "arch/alpha/kernel/setup.c"
 *    Copyright (C) 1995 Linus Torvalds
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 *
 */

/*
 * bootup setup stuff..
 */

#include <linux/config.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/stddef.h>
#include <linux/unistd.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/user.h>
#include <linux/a.out.h>
#include <linux/tty.h>
#include <linux/string.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/major.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/tivoconfig.h>


#include <asm/px_hw.h>
#include <asm/prom.h>
#include <asm/system.h>
#include <asm/pgtable.h>
#include <asm/io.h>
#include <asm/ppcops.h>
#include <asm/ide.h>
#include <asm/px_hw.h>
#include <linux/tivodefs.h>

#ifdef  CONFIG_TCD
/*
 * We need this to compile the generic (that come with the source code)
 * ide drivers.
 */
struct drive_info_struct { char dummy[32]; } drive_info;
#endif

extern int root_mountflags;
static int cpuclock;

void do_unblank_screen(void);
void hard_reset_now(void);
void console_map_init(void);
unsigned long con_init(unsigned long);
void __init init_command_line(void);

#define COMMAND_LINE_SIZE 256
/* static char command_line[COMMAND_LINE_SIZE] = { 0, }; */
char saved_command_line[COMMAND_LINE_SIZE] = { 0, };

static char saved_boot_params[COMMAND_LINE_SIZE];

/*
 * Expect to call in virtual mode, so that we use the normal
 * kernel trap path for machine checks, along with the proper
 * exception table.
 */
#if 0
#if 1
unsigned long 
find_end_of_memory(unsigned long end) {
	extern unsigned long sizemem(unsigned long, int);
	unsigned long v;
	unsigned long iocr = mfspr(SPRN_IOCR);
	mtspr(SPRN_IOCR, iocr | 0x00000400);	/* insure we'll timeout */
	v = sizemem(end, 4096);
	mtspr(SPRN_IOCR, iocr);
	return v;
}
#else
#define find_end_of_memory(x)	(0x80000000 + (8 * 1024 * 1024))
#endif
#endif

unsigned long
find_end_of_memory(unsigned long end) 
{
	extern int PromMemConfig;

#if 0
	switch(PromMemConfig) {
	case 16:
		return(0x80000000 + (16 * 1024 * 1024));
	case 32:
		return(0x80000000 + (32 * 1024 * 1024));
	case 8:
	default:
		break;
	}
	return(0x80000000 + (8 * 1024 * 1024));
#endif
	return(0x80000000 + (PromMemConfig * 1024 * 1024));
}

void __init setup_arch(char **cmdline_p,
	unsigned long * memory_start_p, unsigned long * memory_end_p)
{
	extern char _end[];
	extern unsigned long loops_per_sec;
	extern int _xdebug;
	extern void _xdebug_entry(void);
	void s403_virtualinit(void);
	void s403_print(const char *);
	extern void register_console(void (*)(const char *));
	unsigned int iocr;
	unsigned int br;

	/* HACK: check for first 5 chars == Root=.  If so, rewrite root= and
	   disable console */
	if (!strncmp(saved_boot_params, "Root=", 5)) {
		extern int s403_ready;

		saved_boot_params[0] = 'r';
		s403_ready = -1;
	}
#if 1
	s403_virtualinit();
	register_console(s403_print);
	printk("\rTeleworld Customer Device Starting Up!\n");
#endif
#if 0
{
	void suart_virtualinit(void);
	void suart_print(const char *);


	suart_virtualinit();
	register_console(s403_print);
	printk("\rTeleworld Customer Device Starting Up!\n");
}
#endif
#ifdef CONFIG_XMON
	if (_xdebug & 0x1)
		_xdebug_entry();
#else
	mtspr(SPRN_DBCR, 0x40000000);	/* enter internal debug mode */
#endif
	mtdcr(SPRN_BESR, 0);		/* make sure it's unlocked */

	*memory_start_p = ((unsigned long)_end + PAGE_SIZE)&PAGE_MASK;
	*memory_end_p = find_end_of_memory(*memory_start_p);

	{
	  extern void DmaChInitAll(void);
	  DmaChInitAll();
	}
	
	
	
	/* insure external interrupts are set up properly */
	iocr = mfdcr(SPRN_IOCR);
	iocr &= ~0xffc00000;
#ifdef CONFIG_TCD_EVAL
	iocr |= 0x50000000;
	mtdcr(SPRN_IOCR, iocr);
#else
	iocr |= 0x05400000;
	iocr |= 0x00000400;	/* ignore device timeouts for now */
	mtdcr(SPRN_IOCR, iocr);

	/* special bank register tweaks */

	br = mfdcr(SPRN_BR4);
	br &= ~0xff000000;	/* set up just like uart */
	br |= ((IOBASE_MODEM << 4) & 0xff000000);
	mtdcr(SPRN_BR3, br);
#endif
	/*
	 * Hardcoded NFS boot, for now.
	 */
	/* nfsaddrs=myIP:serverIP:gatewayIP:netmask:clientname rootdir */
#ifdef CONFIG_TCD_EVAL
	strcpy(saved_command_line,
	    "nfsaddrs=192.168.1.23:192.168.1.248:192.168.1.240:255.255.254.0"
	    ":sundance nfsroot=/usr/teleworld/tcdroot/");
#else
/*
	strcpy(saved_command_line,
	    "nfsaddrs=192.168.1.212:192.168.1.248:192.168.1.240:255.255.254.0"
	    ":alfalfa nfsroot=/usr/teleworld/tcdroot/");
*/
	init_command_line();
#endif
	*cmdline_p = saved_command_line;

	/* Set loops_per_sec to a half-way reasonable value,
	   for use until calibrate_delay gets called. */
#ifdef CONFIG_TCD_EVAL
        loops_per_sec = (33 * 2 * 1000000);
#else
        loops_per_sec = (27 * 2 * 1000000);
#endif
}

char *bootpath;
char bootdevice[256];
void *boot_host;
int boot_target;
int boot_part;
kdev_t boot_dev;

int sys_ioperm(unsigned long from, unsigned long num, int on)
{
	return -EIO;
}

int
get_cpuinfo(char *buffer)
{
	int l;
	unsigned int pvr = _get_PVR();

	l = sprintf(buffer, "processor\t: 0\ncpu\t\t: IBM 403GCX\nclock\t\t: %dMHz\n", cpuclock);
	l += sprintf(buffer+l, "revision\t: %d.%d\n",
		     (pvr & 0xff00) >> 8, pvr & 0xff);
	l += sprintf(buffer+l, "bogomips\t: %lu.%02lu\n",
		     (loops_per_sec + 2500) / 500000,
		     (loops_per_sec + 2500) / 5000 % 100);
	l += sprintf(buffer+l, "machine\t\t: Teleworld Customer Device  \n");
	return l;
}

int get_cpuclock(void)
{
  return(cpuclock);
}

void
abort()
{
#ifdef CONFIG_XMON
	extern void xmon(long);
	xmon(0);
#endif
	printk("Abort with no net; entering an infinite loop\n");
	for (;;) {}
}

void do_unblank_screen() {}

void hard_reset_now() {

#define ST551_RESET (1<<1)
#define ST552_RESET (1<<2)
        /* reset the 5505's before going on */
	if (ms2probe()) {
		MsWrite(MS2_SYSTEM_REGISTER,
			MsRead(MS2_SYSTEM_REGISTER) | ST551_RESET);
	}
	if (ms4probe()) {
		MsWrite(MS2_SYSTEM_REGISTER,
			MsRead(MS2_SYSTEM_REGISTER) | ST552_RESET);
	}
	
	/* let's rest for a few milliseconds before rebooting */
	udelay(100000);

	if (ms2probe()) { // if on a combo box

	    char *reg = (char *) 0xFFB40004; // address of modem control register in dss uart

	    *reg = *reg | 0x04; // assert out1 in case we want to use this for hard reset

	    udelay(100000); // give it a chance
	}

	/* cause a full system reset */
	mtspr(SPRN_DBCR, 0x30000000);
}

void console_map_init() {}

unsigned long con_init(unsigned long memstart) {
	/* eventally register on-board serial for bring-up */
	return memstart;
}

#ifndef CONFIG_TCD_EVAL

/* NOTE: this awful hack is valid only for the PX prototype boards. */
/* The ethernet board has a 3 bit id register at the following address. */
/* The 8 bits read contain the id in bits 1-3: XXXX IIIX */
#define PROMADDR 0xfff00000


#define CPUCLOCK_SIGNATURE 0xfeed0000
#define DEFAULT_CPUCLOCK   54
#define DEFAULT_COMBOCLOCK   72
#define MIN_COMBOCLOCK   60

typedef struct {
        int     sig;           /* signature for checked-memory */
        int     size;          /* byte count of checked region */
        char    storage[1];    /* storage area for checked data */
} checkedmem_t;

#define CHECKEDMEM_SIGNATURE 0x426F6E67

char hw_enet_addr[6];
int TiVoBrdType = 0;
checkedmem_t *checkedMem = 0;

int get_videoconfig(void)
{
	nvram_t *conf;
	conf = (nvram_t *)PROMADDR;
	return (conf->videoconfig);
}
EXPORT_SYMBOL(get_videoconfig);

/* Find configured mem size.  If there is a boot parameter "PHYSMEM=NN", use
   that number and ignore the configuration parameter from the PROM.  */
int get_memconfig(void)
{
	nvram_t *conf;
	int memsize = 0;
	int chmemArea = 0;
	char *p;
	void timer_clock_init(int);

	extern unsigned int ms_base_addr;
	TivoConfigValue configValue = kTivoConfigBoardIDClassic;

	p = saved_boot_params;

	/* Look for " PHYSMEM=NN" and " chmem=NNNN"  */
	while (*p) {
	  if (*p++ == ' ') {
			if (!strncmp(p, "PHYSMEM=", 8)) {
				p += 8;
				if (*p >= '0' && *p <= '9')
					memsize = (*p++) - '0';
				if (*p >= '0' && *p <= '9') {
					memsize *= 10;
					memsize += (*p) - '0';
				}
				printk("Memory configured for %dMB\n", memsize);
			} else if (!strncmp(p, "chmem=0x", 8)) {
			       int go = 1;
				p += 8;
				while (go) {
				  if (*p >= '0' && *p <= '9') {
					chmemArea *= 16;
					chmemArea += (*p) - '0';
					p++;
				  } else if (*p >= 'A' && *p <= 'F') {
					chmemArea *= 16;
					chmemArea += (*p) - 'A' + 10;
					p++;
				  } else {
				    go = 0;
				  }
				}
				printk("Checked-mem pool at 0x%X\n", chmemArea);
				checkedMem = (checkedmem_t *) chmemArea;
			}
	  }
	}
	conf = (nvram_t *)PROMADDR;

	if (ms_base_addr) {
		printk("ms_base_addr already initialized.\n");
	}
/* FIXME: beware of that kind of code for future boards >=5 instead of ==*/
	if (conf->videoconfig == 5 || conf->videoconfig == 6) { 
	        configValue = kTivoConfigBoardIDCombo;
		ms_base_addr = NEW_MS_BASE_ADDR;
		printk("New MediaSwitch address map\n");
	} else {
		ms_base_addr = OLD_MS_BASE_ADDR;
		printk("Old MediaSwitch address map\n");
	}

	if (memsize == 0) {
	  if (conf->memconfig < 0)
	    memsize = -conf->memconfig;
	  else
	    memsize = conf->memconfig;
	  if (!memsize)
	    memsize = 8;
	}

	InitTivoConfig( configValue );
	SetTivoConfig( kTivoConfigMemSize, memsize );

	printk("Memory configured for %dMB\n", memsize);
	
	if (!ms4probe()) {
		/* Earlier than P1 Combo box. */
	  cpuclock = DEFAULT_CPUCLOCK;
	} else {
	  /* Combo box... 99% sure it is 72. */
	  cpuclock = conf->cpuclock;
	  if ((cpuclock & 0xffff0000) == CPUCLOCK_SIGNATURE) {
	    cpuclock &= 0xffff;
	    if (!cpuclock)
	      cpuclock = DEFAULT_COMBOCLOCK;
	  } else {
	    cpuclock = DEFAULT_COMBOCLOCK;
	  }
	  if (cpuclock < MIN_COMBOCLOCK)
	    cpuclock = DEFAULT_COMBOCLOCK;
        }
	printk("CPU clock speed: %dMHz\n", cpuclock);
	timer_clock_init(cpuclock);

	return(memsize);
}


void save_boot_params(char *params)
{
	strncpy(saved_boot_params, params, sizeof saved_boot_params);
}

void
init_command_line()
{
	int ip;
	nvram_t *conf;
	char *param;
	unsigned int gwip;

	conf = (nvram_t *)PROMADDR;

	gwip = conf->gateway;

	if (saved_boot_params[0])
		param = saved_boot_params;
	else
		param = conf->bootparam;
	printk("Boot Parameters=%s\n", param);
	memcpy(hw_enet_addr, conf->mac_address, 6);
	printk("Initial MAC address=%x:%x:%x:%x:%x:%x\n", 
		hw_enet_addr[0]&0xff, hw_enet_addr[1]&0xff, 
		hw_enet_addr[2]&0xff, hw_enet_addr[3]&0xff,
		hw_enet_addr[4]&0xff, hw_enet_addr[5]&0xff);
	/* Specify nfs=44:/home/sstone/TiVoRoot */
	/* Expands into nfsaddrs=TCDIPADDR:192.168.1.xx:192.168.1.240:255.255.254.0:hostname /home/sstone/TiVoRoot" */

	/* Skip leading whitespace. */

	while((*param == ' ') || (*param == '\t')) param++;

#define NFSROOTSTR " init=/devbin/nfsinit nfsinit="
#define NFSINITSTR "nfsinit="

	if (!strncmp(NFSINITSTR, param, sizeof(NFSINITSTR)-1)) {
		char *p = param + sizeof(NFSINITSTR) - 1;
		char *hostip;
		char tmpbuf[160];
		char *q, *r;

		r = param + sizeof(NFSINITSTR) - 1;
		strcpy(tmpbuf, NFSROOTSTR);
		q = tmpbuf + sizeof(NFSROOTSTR) - 1;
                while(*r) {
                        *q = *r;
                        if ((*r == ' ') || (*r == '\t'))
                                break;
                        r++; q++;
                }
                *q = 0;

		ip = conf->ipaddr[1].src;

		hostip = p;
		while (*p && (*p != ':')) p++;
		if (*p) *p++ = 0;
		sprintf(saved_command_line, 
			"nfsaddrs=%d.%d.%d.%d:%s:%d.%d.%d.%d:255.255.254.0"
			":%s nfsroot=%s",
			(ip>>24) & 0xff, (ip>>16) & 0xff, (ip>>8) & 0xff,
			ip & 0xff, hostip, 
			(gwip>>24) & 0xff, (gwip>>16) & 0xff, (gwip>>8) & 0xff,
			gwip & 0xff, 
			conf->hostname, p);
		strcat(saved_command_line, tmpbuf);
		printk("Command line: %s\n", saved_command_line);
		return;
	}

	if (!strncmp("nfs=", param, 4)) {
            char *p = param + 4;
            char *hostip;
            ip = conf->ipaddr[1].src;

            hostip = p;
            while (*p && (*p != ':')) p++;
            if (*p) *p++ = 0;
            sprintf(saved_command_line, 
                    "nfsaddrs=%d.%d.%d.%d:192.168.1.%s:%d.%d.%d.%d:255.255.254.0"
                    ":%s nfsroot=%s nfsinit=192.168.1.%s:%s",
                    (ip>>24) & 0xff, (ip>>16) & 0xff, (ip>>8) & 0xff,
                    ip & 0xff, hostip, (gwip>>24) & 0xff, (gwip>>16) & 0xff, 
                    (gwip>>8) & 0xff, gwip & 0xff, conf->hostname, p, hostip, p);
            printk("Command line: %s\n", saved_command_line);
            return;
	}
	if (!strncmp("xnfs=", param, 5)) {
		char *p = param + 5;
		char *hostip;

		ip = conf->ipaddr[1].src;
		

		hostip = p;
		while (*p && (*p != ':')) p++;
		if (*p) *p++ = 0;
		sprintf(saved_command_line, 
			"nfsaddrs=%d.%d.%d.%d:%s:%d.%d.%d.%d:255.255.254.0"
			":%s nfsroot=%s nfsinit=%s:%s",
			(ip>>24) & 0xff, (ip>>16) & 0xff, (ip>>8) & 0xff,
			ip & 0xff, hostip, 
			(gwip>>24) & 0xff, (gwip>>16) & 0xff, (gwip>>8) & 0xff,
			gwip & 0xff,
			conf->hostname, p, hostip, p);
		printk("Command line: %s\n", saved_command_line);
		return;
	}
	if (!strcmp("nfs", param)) {
		ip = conf->ipaddr[1].src;
		sprintf(saved_command_line, 
			"nfsaddrs=%d.%d.%d.%d:192.168.1.44:%d.%d.%d.%d:255.255.254.0"
			":%s nfsroot=/usr/teleworld/tcdroot/ nfsinit=192.168.1.44:/usr/teleworld/tcdroot", 
			(ip>>24) & 0xff,
			(ip>>16) & 0xff,
			(ip>>8) & 0xff,
			ip & 0xff, 
			(gwip>>24) & 0xff, (gwip>>16) & 0xff, (gwip>>8) & 0xff,
			gwip & 0xff, 
			conf->hostname);
		printk("Command line: %s\n", saved_command_line);
		return;
	} 
	// nfs1 and nfs3 are mapped to the new xnfs server.
	if (!strncmp("nfs1", param, 4) || !strncmp("nfs3", param, 4)) {
            char *p = param + 4;
            ip = conf->ipaddr[1].src;
            sprintf(saved_command_line, 
                    "nfsaddrs=%d.%d.%d.%d:192.168.1.250:%d.%d.%d.%d:255.255.254.0"
                    ":%s nfsroot=/ppc/nfsroot/ nfsinit=192.168.1.250:/ppc/nfsroot/ sysgen=true rw %s", 
                    (ip>>24) & 0xff, (ip>>16) & 0xff, (ip>>8) & 0xff,ip & 0xff,
			(gwip>>24) & 0xff, (gwip>>16) & 0xff, (gwip>>8) & 0xff,
			gwip & 0xff, conf->hostname, p);
            printk("Command line: %s\n", saved_command_line);
            return;
        } 
	// nfs2 is the .2 net on xnfs
	if (!strncmp("nfs2", param, 4)) {
            char *p = param + 4;
            ip = conf->ipaddr[1].src;
            sprintf(saved_command_line, 
                    "nfsaddrs=%d.%d.%d.%d:192.168.2.250:%d.%d.%d.%d:255.255.254.0"
                    ":%s nfsroot=/ppc/nfsroot/ nfsinit=192.168.2.250:/ppc/nfsroot sysgen=true rw %s", 
                    (ip>>24) & 0xff, (ip>>16) & 0xff, (ip>>8) & 0xff,ip & 0xff,
			(gwip>>24) & 0xff, (gwip>>16) & 0xff, (gwip>>8) & 0xff,
			gwip & 0xff, conf->hostname, p);
            printk("Command line: %s\n", saved_command_line);
            return;
        } 
	strcpy(saved_command_line, param);
}
#endif

#ifdef CONFIG_CHECKEDMEM

#ifdef CONFIG_BLK_DEV_INITRD
extern unsigned long initrd_start, initrd_end, initrd_below_start_ok;
#endif

unsigned long __init tcd_checkedmem(unsigned long mstart, unsigned long mend)
{
  if (!checkedMem) {
#ifdef DEBUG_CHECKEDMEM
    printk("No checked-memory area\n");
#endif
    return mstart;
  }
  if (checkedMem->sig != CHECKEDMEM_SIGNATURE) {
#ifdef DEBUG_CHECKEDMEM
    printk("No checked-memory signature at 0x%p\n", checkedMem);
#endif
    checkedMem = NULL;
    return mstart;
  }
  if (checkedMem->size + sizeof *checkedMem > mend - mstart) {
#ifdef DEBUG_CHECKEDMEM
    printk("Checked memory area is %ld bytes, MUCH too big!\n",
	   checkedMem->size);
#endif
    checkedMem = NULL;
    return mstart;
  }
#ifdef CONFIG_BLK_DEV_INITRD
#ifdef DEBUG_CHECKEDMEM
  printk("Copying checked memory to 0x%p, %ld bytes\n", checkedMemBase,
	 checkedMem->size);
#endif
  memcpy((void *) mstart, &checkedMem->storage, checkedMem->size);
  initrd_start = mstart;
  mstart += checkedMem->size;
  mstart = (mstart + PAGE_SIZE - 1) & ~PAGE_SIZE;
  initrd_end = mstart;
  initrd_below_start_ok = 1;
  printk("Using checked-memory RAMdisk at 0x%lX size 0x%lx\n",
	 initrd_start, initrd_end - initrd_start);
#endif
  return mstart;
}

#endif



