/*
 * tivoisa.h - definitions for using the ISA adaptor on the TiVo.
 *
 * Paul Mackerras, January 2001.
 */
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/mmu.h>
#include <asm/ppcops.h>

#define USE_PAULS_HACK 0
#define TIVO_ISA_BASE 0xf4000000

#ifndef BR5_TIMING
#define BR5_TIMING 0x40184383
#endif

#undef inb
#undef inb_p
#undef outb
#undef outb_p
#undef outw
#undef outw_p
#undef inw
#undef inw_p

#define u8 unsigned char
#define u16 unsigned short
#define u32 unsigned long

static inline volatile u32 __isa_addr(u32 addr)
{
#if USE_PAULS_HACK
	return (addr & ~1) | ((addr & 1) << 16);
#else
	return addr;
#endif
}

static inline volatile void _setbus16(int set)
{
#if !USE_PAULS_HACK
	if (set) {
		asm volatile("sync");
		mtdcr(SPRN_BR5, BR5_TIMING|0x8000);
		asm volatile("isync");
	} else {
		asm volatile("sync");
		mtdcr(SPRN_BR5, BR5_TIMING);
		asm volatile("isync");
	}
#endif
}

static inline volatile u8 _inb(u32 addr)
{
	addr = __isa_addr(addr);
	addr |= TIVO_ISA_BASE;
	return __inb__((volatile u8 *)addr);
}

static inline volatile u16 _inw(u32 addr)
{
	addr |= TIVO_ISA_BASE;
	return __inw__((volatile u16 *)addr);
}

static inline volatile void _outb(u8 v, u32 addr)
{
	addr = __isa_addr(addr);
	addr |= TIVO_ISA_BASE;
	__outb__(v, (volatile u8 *)addr);
}

static inline volatile void _outw(u16 v, u32 addr)
{
	addr |= TIVO_ISA_BASE;
	__outw__(v, (volatile u16 *)addr);
}


#define inb(port)       _inb((u32)(port))
#define outb(v, port)   _outb(v,(u32)(port))

#define inw(port) _inw((u32)(port))
#define outw(v, port) _outw(v, (u32)(port))

#define inb_p(port)     inb(port)
#define outb_p(v, port) outb(v, port)
#define inw_p(port) inw(port)
#define outw_p(v, port) outw(v, port)


/* this resets the card thanks to some magic paul wired up */
static inline volatile void tivo_isa_reset(void)
{
	outb(1, 0x1);
	udelay(1000);
	outb(1, 0x0);
	udelay(400000);

	/* set to 16 bit byte mode */
	outb(1, 0x4);
	udelay(10000);
}

static inline volatile void set_dcr(void)
{
	u32 iocr;
	asm volatile("sync");
#if USE_PAULS_HACK
	mtdcr(SPRN_BR5, BR5_TIMING | 0x8000);
#else
	mtdcr(SPRN_BR5, BR5_TIMING);
#endif
	iocr = mfdcr(SPRN_IOCR);
	mtdcr(SPRN_IOCR, (iocr & ~(0x0C000000)));
	asm volatile("isync");
}


#if USE_PAULS_HACK
static int set_tlb_entry(void)
{
	int i;
	unsigned long tlbhi, new_tlbhi;
	unsigned long opid = mfspr(SPRN_PID);

	/* Find the existing TLB entry for this area and overwrite it */
	for (i = 0; i < 32; ++i) {
		asm("tlbre %0,%1,0" : "=r" (tlbhi) : "r" (i));
		if ((tlbhi & TLBHI_EPN) != IOBASE_8390)
			continue;
		/* found it! */
		new_tlbhi = (tlbhi & ~TLBHI_PGSZ) | TLBHI_PGSZ_1M;
		if (new_tlbhi != tlbhi) {
			asm("tlbwe %0,%1,0" : : "r" (new_tlbhi), "r" (i));
			asm("isync");
			printk("updated tlb entry %d to 1M\n", i);
		}
		mtspr(SPRN_PID, opid);
		return 0;
	}
	printk("*** didn't find 8390 TLB entry!!\n");
	mtspr(SPRN_PID, opid);
	return -1;
}
#endif


static inline volatile void tivo_isa_setup(void)
{
#if USE_PAULS_HACK
	set_tlb_entry();
#endif
	set_dcr();
	tivo_isa_reset();
}

