#define _AWACS_MIXER_C_

/*
 * sound/awacs_mixer.c
 *
 * Mixer routines for the AWACS cards.
 */
   
#include <linux/config.h>
#include "sound_config.h"

#if defined(CONFIG_AWACS)

#include "awacs_mixer.h"

#include <asm/io.h>
#include "awacs.h"

#undef DEB
#define DEB(what)		/* (what) */
#undef DEB2
#define DEB2(what)		(what)

static int      rec_devices = (SOUND_MASK_MIC);		/* Default recording source */
static int      playthrough_mask = (MASK_MUX_CD | MASK_MUX_AUDIN | MASK_MUX_MIC);

#define POSSIBLE_RECORDING_DEVICES	(SOUND_MASK_SPEAKER | SOUND_MASK_LINE | \
	SOUND_MASK_MIC | SOUND_MASK_CD)

#define SUPPORTED_MIXER_DEVICES		(SOUND_MASK_PCM | SOUND_MASK_SPEAKER | \
	SOUND_MASK_LINE | SOUND_MASK_MIC | SOUND_MASK_CD | SOUND_MASK_VOLUME)

static unsigned short levels[SOUND_MIXER_NRDEVICES] =
{
	0x3232,			/* Master Volume */
	0x3232,			/* Bass */
	0x3232,			/* Treble */
	0x5050,			/* FM */
	0x4b4b,			/* PCM */
	0x3232,			/* PC Speaker */
	0x4b4b,			/* Ext Line */
	0x4b4b,			/* Mic */
	0x4b4b,			/* CD */
	0x6464,			/* Recording monitor */
	0x4b4b,			/* SB PCM */
	0x6464			/* Recording level */
};

static inline void busy_wait(void)
{
	eieio();
	while ((ld_le32(AUD_COD_CONT) & 0x1000000) != 0)
		;
}

static int
awacs_volume_set(int device, int left, int right)
{
int vleft, vright;

	vleft = (left >> 3) & 0xf;		/*   0 <= left < 128   ;   0 <= vleft < 16   */
	vright = (right >> 3) & 0xf;	/*   0 <= right < 128   ;   0 <= vright < 16   */
	
	DEB(printk("awacs_volume_set: left=%d, right=%d;    vleft=%d, vright=%d\n", left, right, vleft, vright));
	
	switch (device)
	{
		case AWACS_MIXVOL_MASTER:
			st_le32(AUD_COD_CONT, MASK_ADDR_VOLSPK | VOLLEFT(vleft) | VOLRIGHT(vright));
			break;

		case AWACS_MIXVOL_SPEAKER:
			st_le32(AUD_COD_CONT, MASK_ADDR_VOLSPK | VOLLEFT(vleft) | VOLRIGHT(vright));
			break;

		case AWACS_MIXVOL_HEADPHONE:
			st_le32(AUD_COD_CONT, MASK_ADDR_VOLHD | VOLLEFT(vleft) | VOLRIGHT(vright));
			break;
	}

	return (left & 0x7f) | ((right & 0x7f) << 8);
}

static int
awacs_gain_set(int device, int left, int right)
{
int vleft, vright;

	vleft = left * 16/100;		/*   0 <= left <= 100   ;   0 <= vleft < 16   */
	vright = right * 16/100;	/*   0 <= right <= 100   ;   0 <= vright < 16   */
	vleft = (vleft > 15) ? 15 : vleft;
	vright = (vright > 15) ? 15 : vright;
	vleft = (vleft < 0) ? 0 : vleft;
	vright = (vright < 0) ? 0 : vright;
	
	DEB(printk("awacs_gain_set: left=%d, right=%d;    vleft=%d, vright=%d\n", left, right, vleft, vright));
	
	switch(device)
	{
		case AWACS_GAIN_LINE:
			if (left == 0 && right == 0)
				playthrough_mask &= ~MASK_MUX_LINE;
			else
				playthrough_mask |= MASK_MUX_LINE;
			
			st_le32(AUD_COD_CONT, MASK_ADDR_MUX | playthrough_mask | \
					GAINRIGHT(vright) | GAINLEFT(vleft) | MASK_GAINLINE);
			break;
			
		case AWACS_GAIN_CD:
			if (left == 0 && right == 0)
				playthrough_mask &= ~MASK_MUX_CD;
			else
				playthrough_mask |= MASK_MUX_CD;
			
			st_le32(AUD_COD_CONT, MASK_ADDR_MUX | playthrough_mask | \
					GAINRIGHT(vright) | GAINLEFT(vleft) | MASK_GAINMIC);
			break;
			
		case AWACS_GAIN_MIC:
			if (left == 0 && right == 0)
				playthrough_mask &= ~MASK_MUX_MIC;
			else
				playthrough_mask |= MASK_MUX_MIC;
			
			st_le32(AUD_COD_CONT, MASK_ADDR_MUX | playthrough_mask | \
					GAINRIGHT(vright) | GAINLEFT(vleft));
			break;
			
	}

	return (left & 0x7f) | ((right & 0x7f) << 8);
}

static int
awacs_mixer_set (int whichDev, unsigned int level)
{
int	left, right, devmask, changed, i;

DEB (printk ("static int awacs_mixer_set(int whichDev = %d, unsigned int level = %x)\n", whichDev, level));

	left = (level) & 0x7f;
	right = (level >> 8) & 0x7f;
	
	switch (whichDev)
	{
		case SOUND_MIXER_VOLUME:	/* Master volume (0-15) */
			levels[whichDev] = awacs_volume_set(AWACS_MIXVOL_MASTER, left, right);
			break;

		case SOUND_MIXER_PCM:	/* AWACS PCM (0-15) */
			levels[whichDev] = awacs_volume_set(AWACS_MIXVOL_MASTER, left, right);
			break;
		case SOUND_MIXER_SPEAKER:	/* PC speaker = internal speaker */
			levels[whichDev] = awacs_volume_set(AWACS_MIXVOL_HEADPHONE, left, right);
			break;
		case SOUND_MIXER_LINE:	/* External line (0-15) */
			levels[whichDev] = awacs_gain_set(AWACS_GAIN_LINE, left, right);
			break;
		case SOUND_MIXER_CD:	/* CD (0-15) */
			levels[whichDev] = awacs_gain_set(AWACS_GAIN_CD, left, right);
			break;
		case SOUND_MIXER_MIC:	/* External microphone (0-15) */
			levels[whichDev] = awacs_gain_set(AWACS_GAIN_MIC, left, right);
			break;


		case SOUND_MIXER_RECSRC:
			devmask = level & POSSIBLE_RECORDING_DEVICES;

			changed = devmask ^ rec_devices;
			rec_devices = devmask;

			for (i = 0; i < SOUND_MIXER_NRDEVICES; i++)
				if (changed & (1 << i))
				{
					awacs_mixer_set (i, levels[i]);
				}
			return rec_devices;
			break;

		default:
			return -(EINVAL);
		}

	return (levels[whichDev]);
}

/*****/

static void
awacs_mixer_reset (void)
{
int		foo;

	DEB (printk ("awacs_mixer.c: void awacs_mixer_reset(void)\n"));

	for (foo = 0; foo < SOUND_MIXER_NRDEVICES; foo++)
		awacs_mixer_set (foo, levels[foo]);
}

int
awacs_mixer_ioctl (int dev, unsigned int cmd, caddr_t arg)
{
	DEB (printk ("awacs_mixer.c: int awacs_mixer_ioctl(unsigned int cmd = %X, unsigned int arg = %X)\n", cmd, arg));

	if (cmd == SOUND_MIXER_PRIVATE1)	/* Unused for now... */
	{
		int level;

		get_user (level, (int *) arg);
	
		DEB (printk("level = %X [int = %X ; char = %X]\n", level, (int) level, (char) level));

		if (level == -1)		/* Return current settings */
		{
			return snd_ioctl_return ((int *) arg, 0);
		}
		else
		{
			return snd_ioctl_return ((int *) arg, 0);
		}
	}


	if (cmd == SOUND_MIXER_PRIVATE2)	/* Unused for now... */
	{
		int level;

		get_user (level, (int *) arg);

		if (level == -1)		/* Return current settings */
		{
			return snd_ioctl_return ((int *) arg, 0);
		}
		else
		{
			return 0;
		}
	}

	if (cmd == SOUND_MIXER_PRIVATE3)	/* Set mute bit */
	{
		int level;

		get_user (level, (int *) arg);

		if (level == -1)		/* Return current settings */
		{
			return snd_ioctl_return ((int *) arg, 0);
		}
		else
		{
			return (0);
		}
	}

	if (((cmd >> 8) & 0xff) == 'M')
	{
		int val;

		if (_IOC_DIR (cmd) & _IOC_WRITE) {
			get_user (val, (int *) arg);
			return snd_ioctl_return ((int *) arg, awacs_mixer_set (cmd & 0xff, val));
		}
		else
		{		/*
				 * Read parameters
				 */

			switch (cmd & 0xff)
			{
				case SOUND_MIXER_RECSRC:
					return snd_ioctl_return ((int *) arg, rec_devices);
					break;
	
				case SOUND_MIXER_STEREODEVS:
					return snd_ioctl_return ((int *) arg, SUPPORTED_MIXER_DEVICES & ~(SOUND_MASK_BASS | SOUND_MASK_TREBLE));
					break;
	
				case SOUND_MIXER_DEVMASK:
					return snd_ioctl_return ((int *) arg, SUPPORTED_MIXER_DEVICES);
					break;
	
				case SOUND_MIXER_RECMASK:
					return snd_ioctl_return ((int *) arg, POSSIBLE_RECORDING_DEVICES & SUPPORTED_MIXER_DEVICES);
					break;
	
				case SOUND_MIXER_CAPS:
					return snd_ioctl_return ((int *) arg, 0);		/* No special capabilities */
					break;
	
				default:
					return snd_ioctl_return ((int *) arg, levels[cmd & 0xff]);
			}
		}
	}
	return -(EINVAL);
}

static struct mixer_operations awacs_mixer_operations =
{
	"AWACS",
	"AWACS chip",
	awacs_mixer_ioctl
};

int
awacs_init_mixer (void)
{
DEB (printk ("awacs_init_mixer: initializing mixer.\n"));

	awacs_mixer_reset ();

	if (num_mixers < MAX_MIXER_DEV)
	{
		audio_devs[awacs_audiodev]->mixer_dev = num_mixers;
		mixer_devs[num_mixers++] = &awacs_mixer_operations;
	}
	return 1;
}

#endif
