/*
 *  linux/arch/ppc/kernel/process.h
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 *
 */

/*
 * This file contains information about how to build and interpret the process
 * log.
 */

#ifdef CONFIG_XMON
#define PLOG_CLKPERUS    (54000000 / 1000000)
#define PLOG_DIFF(cur, prev) \
        (((unsigned long) (cur) - (unsigned long) (prev)) / PLOG_CLKPERUS)

#define PLOG_TYPE_SCE	1
// "System call start"
struct _pl_sce {
    short pid;
    short scnum;
};

#define PLOG_TYPE_SCL	2
// "System call end"
struct _pl_scl {
    int pid;
    int rv;
};

#define PLOG_TYPE_RQ	3
// "Process added to run queue"
struct _pl_rq {
    short pid;
    short priority;
    short rt_priority;
    short policy;
    short pad; // it's happier if we're 4-byte aligned
    char fromintr;
    char preempt;
};

#define PLOG_TYPE_INTRE	4
// "Interrupt start"

#define PLOG_TYPE_INTRL	5
// "Interrupt end"

#define PLOG_TYPE_BHE	6
// "Bottom half start"
#define PLOG_TYPE_BHL	7
// "Bottom half end"

#define PLOG_TYPE_SWE	11
// "Process switch start"
struct _pl_swe {
    short frompid;
    short topid;
};

#define PLOG_TYPE_SWL 12
// "Process switch end"

#define PLOG_TYPE_SPS 13
// "Shared page at 0x%08x swapped out"

#define PLOG_TYPE_FMS 14
// "Data page ot 0x%08x (vaddr=0x%08x) swapped to backing file"

#define PLOG_TYPE_GPS 15
// "Data page at 0x%08x (vaddr=0x%08x) swapped out"

#define PLOG_TYPE_RESWAP 16
// "Data page at 0x%08x (vaddr=0x%08x) tossed out"

#define PLOG_TYPE_PAGEOUT 17
// "Memory at 0x%08x (vaddr=0x%08x) tossed out"

#define PLOG_TYPE_PAGEIN 18
// "Memory at vaddr 0x%08x paged in"

#define PLOG_TYPE_DEMAND_ZERO 19
// "Memory at vaddr 0x%08x demand zeroed"

#define PLOG_TYPE_PAGE_DIRTY 20
// "PTE 0x%08x being marked dirty"

#define PLOG_TYPE_NO_SWAP_PAGE 21
// "No swap page available to swap out pid %d, vaddr 0x%08x"

#define PLOG_TYPE_SHARED_PAGE_IGNORED 22
// "Shared page for pid %d, vaddr 0x%08x multiply referenced, not paged"

#define PLOG_TYPE_DIRTY_MULTIPLE_NO_SWAP 23
// "Page for pid %d, vaddr 0x%08x not swapped out -- it's being DMAed"

#define PLOG_TYPE_I2C 24
// "I2C event type %d, user %d"

#define PLOG_TYPE_DRQ 25
// "marked unrunnable"
struct _pl_drq {
    short pid;
    short priority;
    short rt_priority;
    short policy;
};

#define PLOG_TYPE_WP_PAGE 26
// "Memory at vaddr 0x%08x needs copy-on-write"

#define PLOG_TYPE_WP_PAGE_DONE 27
// "Memory at vaddr 0x%08x copy-on-write complete"

#define PLOG_TYPE_EXIT 28
// "Process exiting with code 0x%x"

#define PLOG_TYPE_PAGEIN_DONE 29
// "Pagein of memory at 0x%08x completed"

#define PLOG_TYPE_AUDIO_LEVEL   30
// "Audio Level: Fifo levels at %d %d %d"
struct _pl_avfifo {
    int audioLevel;
    int videoLevel;
};

#define PLOG_TYPE_VIDEO_LEVEL   31
// "Video Level: Fifo levels at %d %d %d"

#define PLOG_TYPE_LOCK_BT 32
// "Spinlock at addr %08X took %d times.  %08x %08x %08x"

#define PLOG_TYPE_IOCTL 33
// "ioctl( %d %d %d )"
struct _pl_ioctl {
    int pid;
    int fd;
    int cmd;
    int arg;
};

#define PLOG_TYPE_WRITE_STATS 34
// "read stats: %ld calls, %ld writes (%ld, %ld), time %ld (%ld, %ld)"

#define PLOG_OSLINK_READ     100
#define PLOG_OSLINK_WRITE    101
#define PLOG_OSLINK_COMPLETE 102
#define PLOG_OSLINK_TIMEOUT  103
#define PLOG_OSLINK_DEBUG    104
#define PLOG_OSLINK_CAMDONE  105
#define PLOG_OSLINK_HEARTBEAT 106
#define PLOG_OSLINK_BEFORE_VERIFIER 107
#define PLOG_OSLINK_AFTER_VERIFIER  108


#define PLOG_TYPE_UNBLACK_DONE 201
// "The unblack command is complete"

#define PLOG_TYPE_BLACK_DONE   202
// "The black command is complete"

#define PLOG_TYPE_IRCATCH_UP 203
// "Received remote key up"

#define PLOG_TYPE_MOM_TUNE_START 204
// "Starting tuning"

#define PLOG_TYPE_MOM_TUNE_DONE 205
// "Done tuning"

#define PLOG_TYPE_MOM_SEQHDR_RECVED 206
// "Received Seq Hdr"

#define PLOG_TYPE_MOM_TUNE_REQ 207
// "Tune command requested"

#define PLOG_TYPE_MOM_SINK_CC 208
// "Sink performing Channel Change command"

#define PLOG_TYPE_MOM_SINK_UNROLL 209
// "Sink starting playback"

#define PLOG_TYPE_DTUNELIB_CC 210
// "Sending DTune message over Oslink, freq %ld, vscid %lx, ascid %lx"


#define PLOG_TYPE_COBRA_IOCTL 211
// "Receiving Cobra Tuner ioctl %ld (tuner num %ld, fEnter: %ld)"

#define PLOG_TYPE_MOM_SRC_VIDEO_SEG 212
// "MOM source received a video segment, type = %ld"

#define PLOG_TYPE_MOM_SINK_VIDEO_SEG 213
// "MOM Sink writing a video segment, type = %ld, len = %ld"

#define PLOG_TYPE_MOM_SINK_UNBLACK 214
// "MOM Sink was asked to unblack" 

#define PLOG_TYPE_CTXTMGR_RECVD_KEY 215
// "ContextMgr received up key"

#define PLOG_TYPE_VIDEOMGR_TUNE_REQUEST 216
// "VideoMgr received tune request"

#define PLOG_TYPE_WATCHLIVE_CHANNEL_CHANGING_START 217
// "WatchLive told ChannelChanging from video mgr"

#define PLOG_TYPE_WATCHLIVE_CHANNEL_CHANGING_END 218
// "WatchLive told ChannelChanging from video mgr"

#define PLOG_TYPE_WATCHLIVE_CHANNEL_CHANGED 219
// "WatchLive told ChannelChanged from video mgr"

#define PLOG_TYPE_BANNER_QUERY_INFO 220
// "Banner query information returned"

#define PLOG_TYPE_TUNER_STATUS 221
// "Tuner Status is %lx, Tuner 0 = %lx, Tuner 1 = %lx"

#define PLOG_TYPE_CS22_UNBLACK_LATER 222
// "Unblack later ... %ld"

#define PLOG_TYPE_CS22_IOCTL 223
// "CS22 Ioctl : cmd, arg = 0x%lx"

#define PLOG_TYPE_WRITE_AUDIO 224
// "Writing Audio: len = %ld, cBuf = %ld, low water = %ld"

#define PLOG_TYPE_MOM_SINK_NFB 225
// "NFB: err %lx, A %ld:%ld, V %ld:%ld, O %ld:%ld"

#define PLOG_TYPE_MOM_SRC_PFB 226
// "PFB: err %lx, A %ld:%ld, V %ld:%ld, O %ld:%ld"

#define PLOG_TYPE_CS22_INT 227
// "CS22 Interrupt BH: int = 0x%lx"

extern void _plLogEvent(int type, int dataLength, void *data);

extern void _pl_generic_event(int event, int somedata);
extern void _pl_generic_event2(int event, int data1, int data2);
extern void _pl_shared_page_swap(int virtPage);
extern void _pl_filemap_page_swap(int pid, int physPage, int virtPage);
extern void _pl_generic_page_swap(int pid, int physPage, int virtPage);
extern void _pl_reswap(int pid, int physPage, int virtPage);
extern void _pl_pageout(int pid, int physPage, int virtPage);

#endif // CONFIG_XMON
