/*
 *	linux/kernel/softirq.c
 *
 *	Copyright (C) 1992 Linus Torvalds
 *
 * do_bottom_half() runs at normal kernel priority: all interrupts
 * enabled.  do_bottom_half() is atomic with respect to itself: a
 * bottom_half handler need not be re-entrant.
 */

#include <linux/ptrace.h>
#include <linux/errno.h>
#include <linux/kernel_stat.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/mm.h>
#include <linux/smp.h>
#include <linux/smp_lock.h>

#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/bitops.h>

unsigned long intr_count = 0;

int bh_mask_count[32];
unsigned long bh_active = 0;
unsigned long bh_mask = 0;
void (*bh_base[32])(void);


asmlinkage void do_bottom_half(void)
{
    unsigned long active;
    void *bh_active_ptr = &bh_active;
    void (*fn)(void);
    void (**bh)(void);
    int i;

#ifdef CONFIG_XMON
    extern int bh_enter(int);
    int pos;
    int cnt = 0;
#endif

    lock_kernel();
    intr_count=1;
    sti();
    bh = bh_base;
    active = bh_active & bh_mask;

    for( i = 0;  active & (0xFFFFFFFF << i);  ++i )
    {
	if( active & (1 << i) )
	{
	    clear_bit( i, bh_active_ptr );

	    fn = bh[i];

	    if( fn == 0 )
	    {
		printk ("irq.c:bad bottom half entry %08lx\n", 1 << i );
		break;
	    }

#ifdef CONFIG_XMON
            pos = bh_enter(cnt++);
#endif
	    fn();
#ifdef CONFIG_XMON
            bh_leave(pos);
#endif
	}
    }
    cli();
    intr_count=0;
    unlock_kernel();
    return;
}
